#!/usr/bin/env python3
"""
This script loads an asset bundle, removes every object whose original path ID is not between
237426 and 237436, renumbers the allowed objects sequentially (using new IDs 2..N+1 while keeping
the container object as object 1), and rebuilds the main container using the original asset keys.
"""

import os
import sys
import io
from unitypack.asset import Asset
from unitypack.object import ObjectPointer, FFOrderedDict

def update_references(obj, mapping, asset):
    """
    Recursively update any ObjectPointer in obj so that if its path_id is in mapping,
    it is replaced with a new pointer using the new path_id.
    """
    if isinstance(obj, ObjectPointer):
        if obj.path_id in mapping:
            new_ptr = ObjectPointer(None, asset)
            new_ptr.file_id = 0
            new_ptr.path_id = mapping[obj.path_id]
            return new_ptr
        return obj

    if hasattr(obj, "_obj") and isinstance(obj._obj, dict):
        for key, value in obj._obj.items():
            obj._obj[key] = update_references(value, mapping, asset)
        return obj

    if isinstance(obj, dict):
        return { k: update_references(v, mapping, asset) for k, v in obj.items() }
    if isinstance(obj, list):
        return [update_references(item, mapping, asset) for item in obj]
    if isinstance(obj, tuple):
        return tuple(update_references(item, mapping, asset) for item in obj)
    return obj

def rebuild_container_from_orig(orig_container, mapping, asset):
    """
    Rebuild the container from the original container data.
    The original container (an FFOrderedDict) is expected to have:
      - m_Name, m_MainAsset, m_Container, and m_PreloadTable.
    For each entry in m_Container (a tuple of (asset_key, data)),
    if the "asset" pointer's original path_id is in mapping,
    we update that pointer using mapping and preserve the original asset key.
    We then rebuild m_PreloadTable from the new container entries.
    """
    if not isinstance(orig_container, dict):
        return orig_container

    new_container = FFOrderedDict()
    new_container["m_Name"] = orig_container.get("m_Name", "")
    orig_list = orig_container.get("m_Container", [])
    new_list = []
    new_preload = []
    for entry in orig_list:
        if isinstance(entry, tuple) and len(entry) == 2:
            asset_key, data = entry
            ptr = data.get("asset")
            if isinstance(ptr, ObjectPointer) and ptr.path_id in mapping:
                new_ptr = ObjectPointer(None, asset)
                new_ptr.file_id = 0
                new_ptr.path_id = mapping[ptr.path_id]
                data["asset"] = new_ptr
                # Optionally update preloadIndex to reflect new order.
                data["preloadIndex"] = len(new_list)
                new_list.append((asset_key, data))
                new_preload.append(new_ptr)
    new_container["m_Container"] = new_list
    new_container["m_PreloadTable"] = new_preload
    new_container["m_MainAsset"] = orig_container.get("m_MainAsset", 
                                                      FFOrderedDict([("preloadIndex",0), ("preloadSize",0), ("asset", None)]))
    return new_container

def main():
    # Hardcoded path to the original asset bundle.
    input_path = r"C:\Users\wifis\OneDrive\Documents\Desktop\Tessa Fowler\FusionFall-Uploads\WorkSpace\UnityPackFF\CustomAssetBundle-ce09c4c9be8a046ca92e0044f22d1b99"
    base, ext = os.path.splitext(input_path)
    output_path = base + "_new" + ext

    print("Loading asset bundle from:")
    print(input_path)
    # Read entire file into a BytesIO buffer.
    with open(input_path, 'rb') as f:
        asset_data = f.read()
    buf = io.BytesIO(asset_data)
    buf.name = input_path
    asset = Asset.from_file(buf)

    # Determine allowed objects: those with original path IDs between 237426 and 237436.
    allowed_orig_ids = [pid for pid in asset.objects.keys() if 237426 <= pid <= 237436]
    allowed_orig_ids.sort()
    if not allowed_orig_ids:
        print("No allowed objects found!")
        sys.exit(1)

    # Build mapping: assign new path IDs for allowed objects starting at 2.
    mapping = {}
    for i, orig_id in enumerate(allowed_orig_ids):
        mapping[orig_id] = i + 2
    print("Mapping (original -> new):")
    for orig, new in mapping.items():
        print(f"  {orig} -> {new}")

    # Build a new objects dictionary.
    # We assume object 1 is the main container and keep it as is.
    if 1 not in asset.objects:
        print("Main container (object id 1) not found!")
        sys.exit(1)
    new_objects = {}
    new_objects[1] = asset.objects[1]
    for orig_id, new_id in mapping.items():
        obj = asset.objects[orig_id]
        obj.path_id = new_id
        obj._contents = update_references(obj.contents, mapping, asset)
        new_objects[new_id] = obj
    asset._objects = new_objects

    # Update the main container using the original container values.
    container_obj = asset.objects[1]
    if hasattr(container_obj, "contents"):
        orig_container = container_obj.contents
    else:
        orig_container = container_obj._contents

    new_container = rebuild_container_from_orig(orig_container, mapping, asset)
    # Since contents is read-only, update _contents.
    container_obj._contents = new_container

    print("Final new object keys:", list(new_objects.keys()))
    with open(output_path, 'wb') as outf:
        asset.save(outf)
    print(f"Modified asset bundle saved as {output_path}")

if __name__ == '__main__':
    main()
